# Gray Swan AI Python API library

[![PyPI version](https://img.shields.io/pypi/v/grayswan-api.svg)](https://pypi.org/project/grayswan-api/)

The Gray Swan AI Python library provides convenient access to the Gray Swan AI REST API from any Python 3.7+
application. The library includes type definitions for all request params and response fields,
and offers both synchronous and asynchronous clients powered by [httpx](https://github.com/encode/httpx).

It is generated with [Stainless](https://www.stainlessapi.com/).

## Documentation

The REST API documentation can be found [on docs.grayswan.ai](https://docs.grayswan.ai). The full API of this library can be found in [api.md](api.md).

## Installation

```sh
# install from PyPI
pip install --pre grayswan-api
```

## Usage

The full API of this library can be found in [api.md](api.md).

```python
import os
from gray_swan_ai import GraySwanAI

client = GraySwanAI(
    # This is the default and can be omitted
    api_key=os.environ.get("GRAYSWAN_API_TOKEN"),
)

completion_create_response = client.chat.completion.create(
    messages=[
        {
            "role": "system",
            "content": "You are a helpful assistant.",
        },
        {
            "role": "user",
            "content": "I need you to help me with something.",
        },
    ],
    model="cygnet-v0.2",
)
```

While you can provide an `api_key` keyword argument,
we recommend using [python-dotenv](https://pypi.org/project/python-dotenv/)
to add `GRAYSWAN_API_TOKEN="My API Key"` to your `.env` file
so that your API Key is not stored in source control.

## Async usage

Simply import `AsyncGraySwanAI` instead of `GraySwanAI` and use `await` with each API call:

```python
import os
import asyncio
from gray_swan_ai import AsyncGraySwanAI

client = AsyncGraySwanAI(
    # This is the default and can be omitted
    api_key=os.environ.get("GRAYSWAN_API_TOKEN"),
)


async def main() -> None:
    completion_create_response = await client.chat.completion.create(
        messages=[
            {
                "role": "system",
                "content": "You are a helpful assistant.",
            },
            {
                "role": "user",
                "content": "I need you to help me with something.",
            },
        ],
        model="cygnet-v0.2",
    )


asyncio.run(main())
```

Functionality between the synchronous and asynchronous clients is otherwise identical.

## Using types

Nested request parameters are [TypedDicts](https://docs.python.org/3/library/typing.html#typing.TypedDict). Responses are [Pydantic models](https://docs.pydantic.dev) which also provide helper methods for things like:

- Serializing back into JSON, `model.to_json()`
- Converting to a dictionary, `model.to_dict()`

Typed requests and responses provide autocomplete and documentation within your editor. If you would like to see type errors in VS Code to help catch bugs earlier, set `python.analysis.typeCheckingMode` to `basic`.

## Handling errors

When the library is unable to connect to the API (for example, due to network connection problems or a timeout), a subclass of `gray_swan_ai.APIConnectionError` is raised.

When the API returns a non-success status code (that is, 4xx or 5xx
response), a subclass of `gray_swan_ai.APIStatusError` is raised, containing `status_code` and `response` properties.

All errors inherit from `gray_swan_ai.APIError`.

```python
import gray_swan_ai
from gray_swan_ai import GraySwanAI

client = GraySwanAI()

try:
    client.chat.completion.create(
        messages=[
            {
                "role": "system",
                "content": "You are a helpful assistant.",
            },
            {
                "role": "user",
                "content": "I need you to help me with something.",
            },
        ],
        model="cygnet-v0.2",
    )
except gray_swan_ai.APIConnectionError as e:
    print("The server could not be reached")
    print(e.__cause__)  # an underlying Exception, likely raised within httpx.
except gray_swan_ai.RateLimitError as e:
    print("A 429 status code was received; we should back off a bit.")
except gray_swan_ai.APIStatusError as e:
    print("Another non-200-range status code was received")
    print(e.status_code)
    print(e.response)
```

Error codes are as followed:

| Status Code | Error Type                 |
| ----------- | -------------------------- |
| 400         | `BadRequestError`          |
| 401         | `AuthenticationError`      |
| 403         | `PermissionDeniedError`    |
| 404         | `NotFoundError`            |
| 422         | `UnprocessableEntityError` |
| 429         | `RateLimitError`           |
| >=500       | `InternalServerError`      |
| N/A         | `APIConnectionError`       |

### Retries

Certain errors are automatically retried 5 times by default, with a short exponential backoff.
Connection errors (for example, due to a network connectivity problem), 408 Request Timeout, 409 Conflict,
429 Rate Limit, and >=500 Internal errors are all retried by default.

You can use the `max_retries` option to configure or disable retry settings:

```python
from gray_swan_ai import GraySwanAI

# Configure the default for all requests:
client = GraySwanAI(
    # default is 2
    max_retries=0,
)

# Or, configure per-request:
client.with_options(max_retries=5).chat.completion.create(
    messages=[
        {
            "role": "system",
            "content": "You are a helpful assistant.",
        },
        {
            "role": "user",
            "content": "I need you to help me with something.",
        },
    ],
    model="cygnet-v0.2",
)
```

### Timeouts

By default requests time out after 1 minute. You can configure this with a `timeout` option,
which accepts a float or an [`httpx.Timeout`](https://www.python-httpx.org/advanced/#fine-tuning-the-configuration) object:

```python
from gray_swan_ai import GraySwanAI

# Configure the default for all requests:
client = GraySwanAI(
    # 20 seconds (default is 1 minute)
    timeout=20.0,
)

# More granular control:
client = GraySwanAI(
    timeout=httpx.Timeout(60.0, read=5.0, write=10.0, connect=2.0),
)

# Override per-request:
client.with_options(timeout=5.0).chat.completion.create(
    messages=[
        {
            "role": "system",
            "content": "You are a helpful assistant.",
        },
        {
            "role": "user",
            "content": "I need you to help me with something.",
        },
    ],
    model="cygnet-v0.2",
)
```

On timeout, an `APITimeoutError` is thrown.

Note that requests that time out are [retried twice by default](#retries).

## Advanced

### Logging

We use the standard library [`logging`](https://docs.python.org/3/library/logging.html) module.

You can enable logging by setting the environment variable `GRAY_SWAN_AI_LOG` to `debug`.

```shell
$ export GRAY_SWAN_AI_LOG=debug
```

### How to tell whether `None` means `null` or missing

In an API response, a field may be explicitly `null`, or missing entirely; in either case, its value is `None` in this library. You can differentiate the two cases with `.model_fields_set`:

```py
if response.my_field is None:
  if 'my_field' not in response.model_fields_set:
    print('Got json like {}, without a "my_field" key present at all.')
  else:
    print('Got json like {"my_field": null}.')
```

### Accessing raw response data (e.g. headers)

The "raw" Response object can be accessed by prefixing `.with_raw_response.` to any HTTP method call, e.g.,

```py
from gray_swan_ai import GraySwanAI

client = GraySwanAI()
response = client.chat.completion.with_raw_response.create(
    messages=[{
        "role": "system",
        "content": "You are a helpful assistant.",
    }, {
        "role": "user",
        "content": "I need you to help me with something.",
    }],
    model="cygnet-v0.2",
)
print(response.headers.get('X-My-Header'))

completion = response.parse()  # get the object that `chat.completion.create()` would have returned
print(completion)
```

These methods return an [`APIResponse`](https://github.com/grayswanai/grayswan-python/tree/main/src/gray_swan_ai/_response.py) object.

The async client returns an [`AsyncAPIResponse`](https://github.com/grayswanai/grayswan-python/tree/main/src/gray_swan_ai/_response.py) with the same structure, the only difference being `await`able methods for reading the response content.

#### Streaming

Streaming allows you to receive partial results as a completion is generated, so that you do not have to wait for the completion to fully complete before receiving a response.

To stream responses, pass `stream=True` into our client's create method as follows:
```python
completion_create_response = client.chat.completion.create(
    messages=[
        {
            "role": "system",
            "content": "You are a helpful assistant.",
        },
        {
            "role": "user",
            "content": "I need you to help me with something.",
        },
    ],
    model="cygnet-v0.2",
    stream=True
)

for r in completion_create_response:
    delta_content = r.choices[0].delta.content
    print(delta_content)
```

### Making custom/undocumented requests

This library is typed for convenient access to the documented API.

If you need to access undocumented endpoints, params, or response properties, the library can still be used.

#### Undocumented endpoints

To make requests to undocumented endpoints, you can make requests using `client.get`, `client.post`, and other
http verbs. Options on the client will be respected (such as retries) will be respected when making this
request.

```py
import httpx

response = client.post(
    "/foo",
    cast_to=httpx.Response,
    body={"my_param": True},
)

print(response.headers.get("x-foo"))
```

#### Undocumented request params

If you want to explicitly send an extra param, you can do so with the `extra_query`, `extra_body`, and `extra_headers` request
options.

#### Undocumented response properties

To access undocumented response properties, you can access the extra fields like `response.unknown_prop`. You
can also get all the extra fields on the Pydantic model as a dict with
[`response.model_extra`](https://docs.pydantic.dev/latest/api/base_model/#pydantic.BaseModel.model_extra).

### Configuring the HTTP client

You can directly override the [httpx client](https://www.python-httpx.org/api/#client) to customize it for your use case, including:

- Support for proxies
- Custom transports
- Additional [advanced](https://www.python-httpx.org/advanced/clients/) functionality

```python
from gray_swan_ai import GraySwanAI, DefaultHttpxClient

client = GraySwanAI(
    # Or use the `GRAY_SWAN_AI_BASE_URL` env var
    base_url="http://my.test.server.example.com:8083",
    http_client=DefaultHttpxClient(
        proxies="http://my.test.proxy.example.com",
        transport=httpx.HTTPTransport(local_address="0.0.0.0"),
    ),
)
```

### Managing HTTP resources

By default the library closes underlying HTTP connections whenever the client is [garbage collected](https://docs.python.org/3/reference/datamodel.html#object.__del__). You can manually close the client using the `.close()` method if desired, or with a context manager that closes when exiting.

## Versioning

This package generally follows [SemVer](https://semver.org/spec/v2.0.0.html) conventions, though certain backwards-incompatible changes may be released as minor versions:

1. Changes that only affect static types, without breaking runtime behavior.
2. Changes to library internals which are technically public but not intended or documented for external use. _(Please open a GitHub issue to let us know if you are relying on such internals)_.
3. Changes that we do not expect to impact the vast majority of users in practice.

We take backwards-compatibility seriously and work hard to ensure you can rely on a smooth upgrade experience.

We are keen for your feedback; please open an [issue](https://www.github.com/grayswanai/grayswan-python/issues) with questions, bugs, or suggestions.

## Requirements

Python 3.7 or higher.
